from http import HTTPStatus
from logging import Logger
from typing import Any, AnyStr, Dict

from flask import jsonify

from pbench.server import PbenchServerConfig
from pbench.server.api.resources.query_apis import (
    CONTEXT,
    ElasticBase,
    Parameter,
    ParamType,
    PostprocessError,
    Schema,
)


class IndexSearchAPI(ElasticBase):
    """
    Create a search request based on given query parameter.
    """

    def __init__(self, config: PbenchServerConfig, logger: Logger):
        super().__init__(
            config,
            logger,
            Schema(
                Parameter("user", ParamType.USER, required=False),
                Parameter("start", ParamType.DATE, required=True),
                Parameter("end", ParamType.DATE, required=True),
                Parameter("search_term", ParamType.STRING, required=False),
                Parameter("fields", ParamType.LIST, required=False),
            ),
        )

    def flatten_json(self, es_json: JSON) -> JSON:
        """
        Helper function for Postprocess for flattening the json. Any exception raised here will be
        treated as a postprocess error.
        We convert a nested data layer object into a new object with only one layer of key/value pairs.
        Objects such as parent -> child are flattened as parent.child like following:

        parent -> array -> [ele1, ele2] are flattened as
                parent.array.0 = ele1, parent.array.1 = ele2

        parent -> array -> [{key1: value1}, {key2: value2}] are flattened as
                parent.array.key1 = value1, parent.array.key2 = value2

        parent -> array -> [{key1: value1}, {key1: value2}] are flattened as
                parent.array.key1 = [value1, value2]

        parent -> array -> [ele1, ele2, {key1: value1}] are flattened as
                'parent.array.0.0': ele1, 'parent.array.0.1': ele2, 'parent.array.key1': value1

        parent -> array -> [[ele1, ele2], [ele3]] are flattened as
                'parent.array.0.0': ele1, 'parent.array.0.1': ele2, 'parent.array.1.0': ele3

        json_in =
            {
                    "@timestamp": "2021-07-12T14:08:46.563303",
                    "run": {
                        "controller": "dhcp31-171.perf.lab.eng.bos.redhat.com",
                        "name": "pbench-user-benchmark_npalaska-dhcp31-171_2021.07.12T14.08.46"
                    },
                    "host_tools_info": [
                        {
                            "hostname": [{"test1":"dhcp31-171"}, {"test2":"dhcp31-195"}]
                        }
                    ]
                }
        json_out =
            {
                '@timestamp': '2021-07-12T14:08:46.563303',
                'run.controller': 'dhcp31-171.perf.lab.eng.bos.redhat.com',
                'run.name': 'pbench-user-benchmark_npalaska-dhcp31-171_2021.07.12T14.08.46',
                'host_tools_info.hostname.test1': 'dhcp31-171',
                'host_tools_info.hostname.test2': 'dhcp31-195'
            }
        """
        if type(es_json) is not dict:
            raise PostprocessError(
                HTTPStatus.INTERNAL_SERVER_ERROR,
                "Elasticsearch _source value is not a json",
            )

        def recurse(es_json: JSON, prefix: AnyStr, out: JSON) -> JSON:
            if type(es_json) is dict:
                for key in es_json:
                    recurse(es_json[key], prefix + key + ".", out)
            elif type(es_json) is list:
                for i, element in enumerate(es_json):
                    if type(element) is dict:
                        recurse(element, prefix, out)
                    else:
                        recurse(element, prefix + str(i) + ".", out)
            else:
                key = prefix[:-1]
                if key not in out:
                    out[key] = es_json
                elif type(out[key]) == list:
                    out[key].append(es_json)
                else:
                    out[key] = [out[key], es_json]
            return out

        return recurse(es_json, prefix="", out={})

    def assemble(
        self, json_data: Dict[AnyStr, Any], context: CONTEXT
    ) -> Dict[AnyStr, Any]:
        """
        Construct a pbench search query based on a pattern matching given "search_term" parameter
        within a specified date range and which are either owned by a specified username,
        or have been made publicly accessible.

        {
            "user": "username",
            "start": "start-time",
            "end": "end-time"
            "search_term": "string_pattern_to_search"
            "fields": list(interested run document fields returned from IndexMapping API)
        }

        json_data: JSON dictionary of type-normalized parameters
            user: specifies the owner of the data to be searched; it need not
                necessarily be the user represented by the session token
                header, assuming the session user is authorized to view "user"s
                data. If "user": None is specified, then only public datasets
                will be returned.

            "start" and "end" are datetime objects representing a set of Elasticsearch
                run document indices in which to search.

            "search_term": string pattern to match against all the datasets owned by a specified
                    username within a given time range

            "fields": List of run document fields end user is interested in. E.g.
                    ["@timestamp", "run.controller", "run.name", ...]
                    The fields controls what fields are returned by the query, and it has no effect
                    on which fields are searched.
        """
        user = json_data.get("user")
        start = json_data.get("start")
        end = json_data.get("end")

        # If the search_term parameter is not specified we will return all the Elasticsearch hits
        search_term = json_data.get("search_term", "")
        # If no fields are specified, the query will return all the fields from Elasticsearch hits
        selected_fields = json_data.get("fields", [])

        # We need to pass string dates as part of the Elasticsearch query; we
        # use the unconverted strings passed by the caller rather than the
        # adjusted and normalized datetime objects for this.
        start_arg = f"{start:%Y-%m}"
        end_arg = f"{end:%Y-%m}"

        self.logger.info(
            "Search query for user {}, prefix {}: ({} - {}) on query: {}",
            user,
            self.prefix,
            start,
            end,
            search_term,
        )

        uri_fragment = self._gen_month_range("run", start, end)
        self.logger.info("fragment, {}", uri_fragment)
        return {
            "path": f"/{uri_fragment}/_search",
            "kwargs": {
                "json": {
                    "query": {
                        "bool": {
                            "filter": [
                                {"term": self._get_user_term(json_data)},
                                {
                                    "range": {
                                        "@timestamp": {"gte": start_arg, "lte": end_arg}
                                    }
                                },
                                {"query_string": {"query": f"*{search_term}*"}},
                            ]
                        }
                    },
                    "sort": [{"@timestamp": {"order": "desc"}}],
                    "_source": {"include": selected_fields},
                },
                "params": {"ignore_unavailable": "true"},
            },
        }

    def postprocess(
        self, es_json: Dict[AnyStr, Any], context: CONTEXT
    ) -> Dict[AnyStr, Any]:
        """
        Returns a summary of the returned Elasticsearch query results, showing
        the list of dictionaries with user selected fields from request json as keys

        [
            {
                "controller": "alphaville.example.com",
                "results": 2,
                "last_modified_value": 1598473155810.0,
                "last_modified_string": "2020-08-26T20:19:15.810Z"
            }
        ]
        """
        results = []
        # If there are no matches for the user, query, and time range,
        # return the empty list rather than failing.
        try:
            count = es_json["hits"]["total"]["value"]
            if int(count) == 0:
                self.logger.warning("No data returned by Elasticsearch")
                return jsonify(results)
        except KeyError as e:
            raise PostprocessError(
                HTTPStatus.BAD_REQUEST,
                f"Can't find Elasticsearch match data {e} in {es_json!r}",
            )
        except ValueError as e:
            raise PostprocessError(
                HTTPStatus.BAD_REQUEST, f"Elasticsearch hit count {count!r} value: {e}"
            )
        for hit in es_json["hits"]["hits"]:
            s = self.flatten_json(hit["_source"])
            s["id"] = hit["_id"]
            results.append(s)
        # construct response object
        return jsonify(results)

# @pytest.mark.parametrize(
#     "test_json, expected_result",
#     [
#         (
#             {
#                 "@timestamp": "2021-07-12T14:08:46.563303",
#                 "run": {
#                     "controller": "dhcp31-171.perf.lab.eng.bos.redhat.com",
#                     "name": "pbench-user-benchmark_npalaska-dhcp31-171_2021.07.12T14.08.46",
#                 },
#                 "host_tools_info": [
#                     {"hostname": [{"test": "dhcp31-171"}, {"test2": "dhcp31-195"}]}
#                 ],
#             },
#             {
#                 "@timestamp": "2021-07-12T14:08:46.563303",
#                 "run.controller": "dhcp31-171.perf.lab.eng.bos.redhat.com",
#                 "run.name": "pbench-user-benchmark_npalaska-dhcp31-171_2021.07.12T14.08.46",
#                 "host_tools_info.hostname.test": "dhcp31-171",
#                 "host_tools_info.hostname.test2": "dhcp31-195",
#             },
#         ),
#         (
#             {
#                 "a": {"b": 1, "c": None, "d": [False, True, {"c": "test"}]},
#                 "e": "f",
#                 "g": 2.3,
#             },
#             {
#                 "a.b": 1,
#                 "a.c": None,
#                 "a.d.0": False,
#                 "a.d.1": True,
#                 "a.d.c": "test",
#                 "e": "f",
#                 "g": 2.3,
#             },
#         ),
#         ({"": 1}, {"": 1}),
#     ],
# )
# def test_flatten_json(self, client, test_json, expected_result):
#     searchapi = IndexSearch(client.config, client.logger)
#     assert searchapi.flatten_json(test_json) == expected_result