class ConnectionManager(object):
    """
    Represents a simple server connection.
    """

    def __init__(self, base_url, headers={}, timeout=60, verify=True, proxies=None):
        self._base_url = base_url
        self._headers = headers
        self._timeout = timeout
        self._verify = verify
        self._s = requests.Session()
        self._s.auth = lambda x: x  # don't let requests add auth headers

        # retry once to reset connection with Keycloak after  tomcat's ConnectionTimeout
        # see https://github.com/marcospereirampj/python-keycloak/issues/36
        for protocol in ("https://", "http://"):
            adapter = HTTPAdapter(max_retries=1)
            # adds POST to retry whitelist
            allowed_methods = set(adapter.max_retries.allowed_methods)
            allowed_methods.add("POST")
            adapter.max_retries.allowed_methods = frozenset(allowed_methods)

            self._s.mount(protocol, adapter)

        if proxies:
            self._s.proxies.update(proxies)

    def __del__(self):
        self._s.close()

    @property
    def base_url(self):
        """Return base url in use for requests to the server."""
        return self._base_url

    @base_url.setter
    def base_url(self, value):
        """ """
        self._base_url = value

    @property
    def timeout(self):
        """Return timeout in use for request to the server."""
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        """ """
        self._timeout = value

    @property
    def verify(self):
        """Return verify in use for request to the server."""
        return self._verify

    @verify.setter
    def verify(self, value):
        """ """
        self._verify = value

    @property
    def headers(self):
        """Return header request to the server."""
        return self._headers

    @headers.setter
    def headers(self, value):
        """ """
        self._headers = value

    def param_headers(self, key):
        """
        Return a specific header parameter.
        :param key: (str) Header parameters key.
        :returns: If the header parameters exist, return its value.
        """
        return self.headers.get(key)

    def clean_headers(self):
        """Clear header parameters."""
        self.headers = {}

    def exist_param_headers(self, key):
        """Check if the parameter exists in the header.
        :param key: (str) Header parameters key.
        :returns: If the header parameters exist, return True.
        """
        return self.param_headers(key) is not None

    def add_param_headers(self, key, value):
        """Add a single parameter inside the header.
        :param key: (str) Header parameters key.
        :param value: (str) Value to be added.
        """
        self.headers[key] = value

    def del_param_headers(self, key):
        """Remove a specific parameter.
        :param key: (str) Key of the header parameters.
        """
        self.headers.pop(key, None)

    def raw_get(self, path, **kwargs):
        """Submit get request to the path.
        :param path: (str) Path for request.
        :returns: Response the request.
        :raises: HttpError Can't connect to server.
        """

        try:
            return self._s.get(
                urljoin(self.base_url, path),
                params=kwargs,
                headers=self.headers,
                timeout=self.timeout,
                verify=self.verify,
            )
        except Exception as e:
            raise KeycloakConnectionError("Can't connect to server (%s)" % e)

    def raw_post(self, path, data, **kwargs):
        """Submit post request to the path.
        :param path: (str) Path for request.
        :param data: (dict) Payload for request.
        :returns: Response the request.
        :raises: HttpError Can't connect to server.
        """
        try:
            return self._s.post(
                urljoin(self.base_url, path),
                params=kwargs,
                data=data,
                headers=self.headers,
                timeout=self.timeout,
                verify=self.verify,
            )
        except Exception as e:
            raise KeycloakConnectionError("Can't connect to server (%s)" % e)

    def raw_put(self, path, data, **kwargs):
        """Submit put request to the path.
        :param path: (str) Path for request.
        :param data: (dict) Payload for request.
        :returns: Response the request.
        :raises: HttpError Can't connect to server.
        """
        try:
            return self._s.put(
                urljoin(self.base_url, path),
                params=kwargs,
                data=data,
                headers=self.headers,
                timeout=self.timeout,
                verify=self.verify,
            )
        except Exception as e:
            raise KeycloakConnectionError("Can't connect to server (%s)" % e)

    def raw_delete(self, path, data={}, **kwargs):
        """Submit delete request to the path.
        :param path: (str) Path for request.
        :param data: (dict) Payload for request.
        :returns: Response the request.
        :raises: HttpError Can't connect to server.
        """
        try:
            return self._s.delete(
                urljoin(self.base_url, path),
                params=kwargs,
                data=data,
                headers=self.headers,
                timeout=self.timeout,
                verify=self.verify,
            )
        except Exception as e:
            raise KeycloakConnectionError("Can't connect to server (%s)" % e)


    # def get_client_role(self, client_id: str, role_name: str):
    #     """Get client role id by name.
    #     :param client_id: id of client (not client-id)
    #     :param role_name: role name (not id!)
    #     RoleRepresentation
    #     https://www.keycloak.org/docs-api/18.0/rest-api/index.html#_rolerepresentation
    #     :return: role_id
    #     """
    #     params_path = {"realm-name": self.user_realm, "id": client_id, "role-name": role_name}
    #     return self._get(URL_ADMIN_CLIENT_ROLE.format(**params_path)).json()
    #
    # def create_client_role(self, client_role_id, payload):
    #     """Create a client role.
    #     RoleRepresentation
    #     https://www.keycloak.org/docs-api/18.0/rest-api/index.html#_rolerepresentation
    #     :param client_role_id: id of client
    #     :param payload: RoleRepresentation
    #     :return: Client role name
    #     """
    #     params_path = {"realm-name": self.user_realm, "id": client_role_id}
    #     data_raw = self._post(
    #         URL_ADMIN_CLIENT_ROLES.format(**params_path), data=json.dumps(payload)
    #     )
    #     raise_error_from_response(
    #         data_raw, KeycloakPostError, expected_codes=[201], skip_exists=skip_exists
    #     )
    #     _last_slash_idx = data_raw.headers["Location"].rindex("/")
    #     return data_raw.headers["Location"][_last_slash_idx + 1:]  # noqa: E203
    #
    # def update_client_role(self, client_role_id, role_name, payload):
    #     """Update a client role.
    #     RoleRepresentation
    #     https://www.keycloak.org/docs-api/18.0/rest-api/index.html#_rolerepresentation
    #     :param client_role_id: id of client (not client-id)
    #     :param role_name: role's name (not id!)
    #     :param payload: RoleRepresentation
    #     """
    #     params_path = {"realm-name": self.realm_name, "id": client_role_id, "role-name": role_name}
    #     data_raw = self.raw_put(
    #         urls_patterns.URL_ADMIN_CLIENT_ROLE.format(**params_path), data=json.dumps(payload)
    #     )
    #     return raise_error_from_response(data_raw, KeycloakPutError, expected_codes=[204])
    #
    # def delete_client_role(self, client_role_id, role_name):
    #     """Delete a client role.
    #     RoleRepresentation
    #     https://www.keycloak.org/docs-api/18.0/rest-api/index.html#_rolerepresentation
    #     :param client_role_id: id of client (not client-id)
    #     :param role_name: role's name (not id!)
    #     """
    #     params_path = {"realm-name": self.realm_name, "id": client_role_id, "role-name": role_name}
    #     data_raw = self.raw_delete(urls_patterns.URL_ADMIN_CLIENT_ROLE.format(**params_path))
    #     return raise_error_from_response(data_raw, KeycloakDeleteError, expected_codes=[204])
    #
    # def assign_client_role(self, user_id, client_id, roles):
    #     """Assign a client role to a user.
    #     :param user_id: id of user
    #     :param client_id: id of client
    #     :param roles: roles list or role (use RoleRepresentation)
    #     :return: Keycloak server response
    #     """
    #     payload = roles if isinstance(roles, list) else [roles]
    #     params_path = {"realm-name": self.realm_name, "id": user_id, "client-id": client_id}
    #     return self._post(URL_ADMIN_USER_CLIENT_ROLES.format(**params_path), json=payload).json()